/*
 * Copyright 2013 Freescale Semiconductor
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*/

#include "libmcc2.h"
#include <sys/ioctl.h>
#include <errno.h>
#include <stddef.h>

#include <linux/mcc_common.h>
#include <linux/mcc_config_linux.h>

#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/mman.h>

// local, private data
static int fd = -1;	// the file descriptor used throughout
static unsigned int bookeeping_p;

static MCC_CORE this_core = MCC_CORE_NUMBER;
static MCC_NODE this_node;
static MCC_ENDPOINT send_endpoint, recv_endpoint;
static unsigned int current_timeout_us = 0xFFFFFFFF; //0

#define MCC_ENDPOINT_IN_USE(e) (e.port != MCC_RESERVED_PORT_NUMBER)

/*!
 * \brief This function initializes the Multi Core Communication subsystem for a given node.
 *
 * This should only be called once per node (once in MQX, once per process in Linux).
 *
 * \param[in] node Node number that will be used in endpoints created by this process.
 *
 * \return MCC_SUCCESS
 * \return MCC_ERR_DEV (cant open /dev/mcc - kernel module not loaded)
 * \return MCC_ERR_NOMEM (Cant map shared memory)
*/
int mcc_initialize(MCC_NODE node)
{
	int i;

	this_node = node;

	send_endpoint.port = MCC_RESERVED_PORT_NUMBER;
	recv_endpoint.port = MCC_RESERVED_PORT_NUMBER;

	fd = open("/dev/mcc", O_RDWR);
	if(fd < 0)
		return MCC_ERR_DEV;

	if(ioctl(fd, MCC_SET_NODE, &node) < 0)
	{
		if(errno == EFAULT)
			return MCC_ERR_DEV;
	}

	bookeeping_p = (unsigned int) mmap(0, sizeof(MCC_BOOKEEPING_STRUCT), PROT_WRITE | PROT_READ, MAP_SHARED, fd, 0);
	if(bookeeping_p < 0)
		return MCC_ERR_NOMEM;

    return MCC_SUCCESS;
}

/*!
 * \brief This function de-initializes the Multi Core Communication subsystem for a given node.
 *
 * Frees all resources of the node. Deletes all endpoints and frees any buffers that may have been queued there.
 *
 * \param[in] node Node number to be deinitialized.
 *
 * \return MCC_SUCCESS
 * \return MCC_ERR_DEV (internal driver error)
 * \return MCC_ERR_ENDPOINT (invalid endpoint)
 * \return MCC_ERR_SEMAPHORE (semaphore handling error)
 */
int mcc_destroy(MCC_NODE node)
{
    close(fd);

    return MCC_SUCCESS;
}

/*!
 * \brief This function creates an endpoint.
 *
 * Create an endpoint on the local node with the specified port number.
 * The core and node provided in endpoint must match the callers core and
 * node and the port argument must match the endpoint port.
 *
 * \param[out] endpoint Pointer to the endpoint triplet to be created.
 * \param[in] port Port number.
 *
 * \return MCC_SUCCESS
 * \return MCC_ERR_DEV (internal driver error)
 * \return MCC_ERR_NOMEM (maximum number of endpoints exceeded)
 * \return MCC_ERR_ENDPOINT (invalid value for core, node, or port)
 * \return MCC_ERR_SEMAPHORE (semaphore handling error)
 */
int mcc_create_endpoint(MCC_ENDPOINT *endpoint, MCC_PORT port)
{
	endpoint->port = port;
	if(ioctl(fd, MCC_CREATE_ENDPOINT, &endpoint) < 0)
	{
		if(errno == EFAULT)
			return MCC_ERR_DEV;
		if(errno == EBUSY)
			return MCC_ERR_SEMAPHORE;
		if(errno == EINVAL)
			return MCC_ERR_ENDPOINT;
		if(errno == ENOMEM)
			return MCC_ERR_NOMEM;
	}

    return MCC_SUCCESS;
}

/*!
  * \brief This function destroys an endpoint.
 *
 * Destroy an endpoint on the local node and frees any buffers that may be queued.
 *
 * \param[in] endpoint Pointer to the endpoint triplet to be deleted.
 *
 * \return MCC_SUCCESS
 * \return MCC_ERR_DEV (internal driver error)
 * \return MCC_ERR_ENDPOINT (the endpoint doesn't exist)
 * \return MCC_ERR_SEMAPHORE (semaphore handling error)
*/
int mcc_destroy_endpoint(MCC_ENDPOINT *endpoint)
{
	if(endpoint->node != this_node)
		return MCC_ERR_ENDPOINT;

	if(ioctl(fd, MCC_DESTROY_ENDPOINT, &endpoint) < 0)
	{
		if(errno == EFAULT)
			return MCC_ERR_DEV;
		if(errno == EBUSY)
			return MCC_ERR_SEMAPHORE;
		if(errno == EINVAL)
			return MCC_ERR_ENDPOINT;
	}

    return MCC_SUCCESS;
}

/*!
 * \brief This function sends a message to an endpoint.
 *
 * The message is copied into the MCC buffer and the destination core is signaled.
 *
 * \param[in] endpoint Pointer to the receiving endpoint to send to.
 * \param[in] msg Pointer to the message to be sent.
 * \param[in] msg_size Size of the message to be sent in bytes.
 * \param[in] timeout_us Timeout, in microseconds, to wait for a free buffer. A value of 0 means dont wait (non-blocking call), 0xffffffff means wait forever (blocking call).
 *
 * \return MCC_SUCCESS
 * \return MCC_ERR_DEV (internal driver error)
 * \return MCC_ERR_ENDPOINT (the endpoint does not exist)
 * \return MCC_ERR_SEMAPHORE (semaphore handling error)
 * \return MCC_ERR_INVAL (an invalid address has been supplied for msg or the msg_size exceeds the size of a data buffer)
 * \return MCC_ERR_TIMEOUT (timeout exceeded before a buffer became available)
 * \return MCC_ERR_NOMEM (no free buffer available and timeout_us set to 0)
 * \return MCC_ERR_SQ_FULL (signal queue is full)
 */
int mcc_send(MCC_ENDPOINT *from, MCC_ENDPOINT *endpoint, void *msg, MCC_MEM_SIZE msg_size, unsigned int timeout_ms)
{
	struct mcc_send_info_struct sis;
	errno = 0;

	if(msg_size > MCC_ATTR_BUFFER_SIZE_IN_BYTES)
		return MCC_ERR_INVAL;

	sis.src_endpoint = from;
	sis.dest_endpoint = endpoint;
	sis.msg = msg;
	sis.msg_size = msg_size;
	sis.timeout_ms = timeout_ms;

	ioctl(fd, MCC_SEND, &sis);

	switch (errno)
	{
	case EFAULT:
		return MCC_ERR_DEV;
	case EBUSY:
		return MCC_ERR_SEMAPHORE;
	case EINVAL:
		return MCC_ERR_ENDPOINT;
	case ETIME:
		return MCC_ERR_TIMEOUT;
	case ENOMEM:
		return MCC_ERR_NOMEM;
	case MCC_SUCCESS:
		return MCC_SUCCESS;
	}
	return MCC_ERR_DEV;
}

int mcc_send_nocopy(MCC_ENDPOINT *src_endpoint, MCC_ENDPOINT *dest_endpoint, void *buffer_p, MCC_MEM_SIZE msg_size)
{
        struct mcc_send_nocopy_info_struct snis;
        errno = 0;

        if(msg_size > MCC_ATTR_BUFFER_SIZE_IN_BYTES)
                return MCC_ERR_INVAL;

        snis.src_endpoint = src_endpoint;
        snis.dest_endpoint = dest_endpoint;
        snis.offset = (unsigned int)buffer_p - bookeeping_p;
        snis.msg_size = msg_size;

        ioctl(fd, MCC_SEND_NOCOPY, &snis);
        switch (errno)
        {
        case EFAULT:
                return MCC_ERR_DEV;
        case EBUSY:
                return MCC_ERR_SEMAPHORE;
        case EINVAL:
                return MCC_ERR_ENDPOINT;
        case ETIME:
                return MCC_ERR_TIMEOUT;
        case ENOMEM:
                return MCC_ERR_NOMEM;
        case MCC_SUCCESS:
                return MCC_SUCCESS;
        }
        return MCC_ERR_DEV;
}


/*!
* \brief This function receives a message from the specified endpoint if one is available.
 *        The data will be copied from the receive buffer into the user supplied buffer.
 *
 * This is the "receive with copy" version of the MCC receive function. This version is simple
 * to use but includes the cost of copying the data from shared memory into the user space buffer.
 * The user has no obligation or burden to manage shared memory buffers.
 *
 * \param[in] endpoint Pointer to the receiving endpoint to receive from.
 * \param[in] buffer Pointer to the user-app. buffer data will be copied into.
 * \param[in] buffer_size The maximum number of bytes to copy.
 * \param[out] recv_size Pointer to an MCC_MEM_SIZE that will contain the number of bytes actually copied into the buffer.
 * \param[in] timeout_us Timeout, in microseconds, to wait for a free buffer. A value of 0 means dont wait (non-blocking call), 0xffffffff means wait forever (blocking call).
 *
 * \return MCC_SUCCESS
 * \return MCC_ERR_DEV (internal driver error)
 * \return MCC_ERR_ENDPOINT (the endpoint does not exist)
 * \return MCC_ERR_SEMAPHORE (semaphore handling error)
 * \return MCC_ERR_TIMEOUT (timeout exceeded before a buffer became available)
 */
int mcc_recv(MCC_ENDPOINT *from, MCC_ENDPOINT *endpoint, void *buffer, MCC_MEM_SIZE buffer_size, MCC_MEM_SIZE *recv_size, unsigned int timeout_ms)
{
	struct mcc_recv_info_struct ris;

	ris.src_endpoint = from;
	ris.dest_endpoint = endpoint;
	ris.buffer = buffer;
	ris.buffer_size = buffer_size;
	ris.recv_size = recv_size;
	ris.timeout_ms = timeout_ms;

	return ioctl(fd, MCC_RECV, &ris);
}

/*!
 * \brief This function receives a message from the specified endpoint if one is available. The data is NOT copied into the user-app. buffer.
 *
 * This is the "zero-copy receive" version of the MCC receive function. No data is copied, just
 * the pointer to the data is returned. This version is fast, but requires the user to manage
 * buffer allocation. Specifically the user must decide when a buffer is no longer in use and
 * make the appropriate API call to free it.
 *
 * \param[in] endpoint Pointer to the receiving endpoint to receive from.
 * \param[out] buffer_p Pointer to the MCC buffer of the shared memory where the received data is stored.
 * \param[out] recv_size Pointer to an MCC_MEM_SIZE that will contain the number of valid bytes in the buffer.
 * \param[in] timeout_us Timeout, in microseconds, to wait for a free buffer. A value of 0 means dont wait (non-blocking call), 0xffffffff means wait forever (blocking call).
 *
 * \return MCC_SUCCESS
 * \return MCC_ERR_DEV (internal driver error)
 * \return MCC_ERR_ENDPOINT (the endpoint does not exist)
 * \return MCC_ERR_SEMAPHORE (semaphore handling error)
 * \return MCC_ERR_TIMEOUT (timeout exceeded before a buffer became available)
 */
int mcc_recv_nocopy(MCC_ENDPOINT *from, MCC_ENDPOINT *endpoint, void **buffer_p, MCC_MEM_SIZE *recv_size, unsigned int timeout_ms)
{
	struct mcc_recv_nocopy_info_struct rnis;
	errno = 0;

	rnis.src_endpoint = from;
	rnis.dest_endpoint = endpoint;
	rnis.recv_size = recv_size;
	rnis.timeout_ms = timeout_ms;

	ioctl(fd, MCC_RECV_NOCOPY, &rnis);
	*buffer_p = (void *)(rnis.offset + bookeeping_p);

	switch (errno)
        {
        case EFAULT:
                return MCC_ERR_DEV;
        case EBUSY:
                return MCC_ERR_SEMAPHORE;
        case EINVAL:
                return MCC_ERR_ENDPOINT;
        case ETIME:
                return MCC_ERR_TIMEOUT;
        case ENOMEM:
                return MCC_ERR_NOMEM;
        case 0:
                return MCC_SUCCESS;
        }
        return MCC_ERR_DEV;
}

/*!
 * \brief This function returns number of buffers currently queued at the endpoint.
 *
 * Checks if messages are available on a receive endpoint. The call only checks the
 * availability of messages and does not dequeue them.
 *
 * \param[in] endpoint Pointer to the endpoint structure.
 * \param[out] num_msgs Pointer to an int that will contain the number of buffers queued.
 *
 * \return MCC_SUCCESS
 * \return MCC_ERR_ENDPOINT (the endpoint does not exist)
 * \return MCC_ERR_SEMAPHORE (semaphore handling error)
 */

int mcc_msgs_available(MCC_ENDPOINT *endpoint, unsigned int *num_msgs)
{
	struct mcc_queue_info_struct q_info;
	q_info.endpoint.core = endpoint->core;
	q_info.endpoint.node = endpoint->node;
	q_info.endpoint.port = endpoint->port;

	if(ioctl(fd, MCC_GET_QUEUE_INFO, &q_info) < 0)
		return MCC_ERR_ENDPOINT;

	*num_msgs = q_info.current_queue_length;

	return MCC_SUCCESS;
}

/*!
 * \brief This function dequeues a buffer from the free list.
 *
 * The application has take the responsibility for MCC buffer de-allocation and
 * filling the data to be sent into the pre-allocated MCC buffer.
 *
 * \param[out] buffer Pointer to the MCC buffer dequeued from the free list.
 * \param[out] buf_size Pointer to an MCC_MEM_SIZE that is used for passing the size of the dequeued MCC buffer to the application.
 * \param[in] timeout_ms Timeout, in milliseconds, to wait for a free buffer. A value of 0 means don't wait (non-blocking call). A value of 0xffffffff means wait forever (blocking call).
 *
 * \return MCC_SUCCESS
 * \return MCC_ERR_SEMAPHORE (semaphore handling error)
 * \return MCC_ERR_TIMEOUT (timeout exceeded before a buffer became available)
 * \return MCC_ERR_NOMEM (no free buffer available and timeout_ms set to 0)
 *
 * \see mcc_send_nocopy
 */
int mcc_get_buffer(void **buffer, MCC_MEM_SIZE *buf_size, unsigned int timeout_ms)
{
	struct mcc_get_buffer_struct gbs;
	gbs.buf_size = buf_size;
	gbs.timeout_ms = timeout_ms;

	if(ioctl(fd, MCC_GET_RECEIVE_BUFFER, &gbs))
		return MCC_ERR_INVAL;
	*buffer = (void *)(gbs.offset + bookeeping_p);

	return MCC_SUCCESS;
}

/*!
  * \brief This function frees a buffer previously returned by mcc_recv_nocopy().
 *
 * Once the zero-copy mechanism of receiving data is used this function
 * has to be called to free a buffer and to make it available for the next data
 * transfer.
 *
 * \param[in] buffer Pointer to the buffer to be freed.
 *
 * \return MCC_SUCCESS
 * \return MCC_ERR_ENDPOINT (the endpoint does not exist)
 * \return MCC_ERR_SEMAPHORE (semaphore handling error)
 */
int mcc_free_buffer(void *buffer)
{
	unsigned int offset = (unsigned int)buffer - bookeeping_p;

	if(ioctl(fd, MCC_FREE_RECEIVE_BUFFER, &offset) < 0)
		return MCC_ERR_INVAL;

	return MCC_SUCCESS;
}

/*!
  * \brief This function returns information about the MCC sub system.
 *
 * Returns implementation specific information.
 *
 * \param[in] node Node number.
 * \param[out] info_data Pointer to the MCC_INFO_STRUCT structure to hold returned data.
 *
 * \return MCC_SUCCESS
 * \return MCC_ERR_DEV (internal driver error)
 * \return MCC_ERR_SEMAPHORE (semaphore handling error)
 */
int mcc_get_info(MCC_NODE node, MCC_INFO_STRUCT* info_data)
{
	if(ioctl(fd, MCC_GET_INFO, &info_data) < 0)
	{
		if(errno == EFAULT)
			return MCC_ERR_DEV;
		if(errno == EBUSY)
			return MCC_ERR_SEMAPHORE;
		if(errno == EINVAL)
			return MCC_ERR_ENDPOINT;
	}
	return MCC_SUCCESS;
}

